@@ -20,7 +20,18 @@
 
 
 __appname__ = "hy"
-__version__ = "0.9.5"
+__version__ = "0.9.6"
+
+
+from hy.models.expression import HyExpression  # NOQA
+from hy.models.integer import HyInteger  # NOQA
+from hy.models.keyword import HyKeyword  # NOQA
+from hy.models.complex import HyComplex  # NOQA
+from hy.models.string import HyString  # NOQA
+from hy.models.symbol import HySymbol  # NOQA
+from hy.models.float import HyFloat  # NOQA
+from hy.models.dict import HyDict  # NOQA
+from hy.models.list import HyList  # NOQA
 
 
 import hy.importer  # NOQA
