@@ -18,8 +18,9 @@
 # FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 # DEALINGS IN THE SOFTWARE.
 
-from hy.compiler import hy_compile
 from py_compile import wr_long, MAGIC
+from hy.compiler import hy_compile
+from hy.models import HyObject
 from hy.core import process
 from hy.lex import tokenize
 
@@ -28,13 +29,21 @@ from io import open
 import marshal
 import imp
 import sys
+import ast
 import os
-
+import __future__
 
 if sys.version_info[0] >= 3:
     from io import StringIO
+    long_type = int
 else:
     from StringIO import StringIO  # NOQA
+    import __builtin__
+    long_type = long  # NOQA
+
+
+def compile_(ast, filename, mode):
+    return compile(ast, filename, mode, __future__.CO_FUTURE_DIVISION)
 
 
 def import_buffer_to_hst(fd):
@@ -49,38 +58,60 @@ def import_file_to_hst(fpath):
 
 def import_file_to_ast(fpath):
     tree = import_file_to_hst(fpath)
-    ast = hy_compile(tree)
-    return ast
+    _ast = hy_compile(tree)
+    return _ast
 
 
 def import_string_to_ast(buff):
     tree = import_buffer_to_hst(StringIO(buff))
-    ast = hy_compile(tree)
-    return ast
+    _ast = hy_compile(tree)
+    return _ast
 
 
 def import_file_to_module(name, fpath):
-    ast = import_file_to_ast(fpath)
+    _ast = import_file_to_ast(fpath)
     mod = imp.new_module(name)
     mod.__file__ = fpath
-    eval(compile(ast, fpath, "exec"), mod.__dict__)
+    eval(compile_(_ast, fpath, "exec"), mod.__dict__)
     return mod
 
 
+def hy_eval(hytree, namespace):
+    foo = HyObject()
+    foo.start_line = 0
+    foo.end_line = 0
+    foo.start_column = 0
+    foo.end_column = 0
+    hytree.replace(foo)
+    _ast = hy_compile(hytree, root=ast.Expression)
+    return eval(compile_(_ast, "<eval>", "eval"), namespace)
+
+
 def write_hy_as_pyc(fname):
     with open(fname, 'U') as f:
         try:
-            timestamp = long(os.fstat(f.fileno()).st_mtime)
+            st = os.fstat(f.fileno())
         except AttributeError:
-            timestamp = long(os.stat(fname).st_mtime)
+            st = os.stat(fname)
+        timestamp = long_type(st.st_mtime)
 
     _ast = import_file_to_ast(fname)
-    code = compile(_ast, fname, "exec")
+    code = compile_(_ast, fname, "exec")
     cfile = "%s.pyc" % fname[:-len(".hy")]
 
-    with open(cfile, 'wb') as fc:
-        fc.write('\0\0\0\0')
+    if sys.version_info[0] >= 3:
+        open_ = open
+    else:
+        open_ = __builtin__.open
+
+    with open_(cfile, 'wb') as fc:
+        if sys.version_info[0] >= 3:
+            fc.write(b'\0\0\0\0')
+        else:
+            fc.write('\0\0\0\0')
         wr_long(fc, timestamp)
+        if (sys.version_info[0] >= 3 and sys.version_info[1] >= 3):
+            wr_long(fc, st.st_size)
         marshal.dump(code, fc)
         fc.flush()
         fc.seek(0, 0)
