@@ -20,8 +20,11 @@
 
 from hy.models.expression import HyExpression
 from hy.models.integer import HyInteger
+from hy.models.float import HyFloat
+from hy.models.complex import HyComplex
 from hy.models.symbol import HySymbol
 from hy.models.string import HyString
+from hy.models.keyword import HyKeyword
 from hy.models.dict import HyDict
 from hy.models.list import HyList
 
@@ -45,6 +48,8 @@ def _resolve_atom(obj):
     Resolve a bare atom into one of the following (in order):
 
         - Integer
+        - Float
+        - Complex
         - Symbol
     """
     try:
@@ -52,6 +57,16 @@ def _resolve_atom(obj):
     except ValueError:
         pass
 
+    try:
+        return HyFloat(obj)
+    except ValueError:
+        pass
+
+    try:
+        return HyComplex(obj)
+    except ValueError:
+        pass
+
     table = {
         "true": "True",
         "false": "False",
@@ -61,7 +76,10 @@ def _resolve_atom(obj):
     if obj in table:
         return HySymbol(table[obj])
 
-    if obj.startswith("*") and obj.endswith("*") and obj != "*":
+    if obj.startswith(":"):
+        return HyKeyword(obj)
+
+    if obj.startswith("*") and obj.endswith("*") and obj not in ("*", "**"):
         obj = obj[1:-1].upper()
 
     if "-" in obj and obj != "-":
@@ -255,6 +273,46 @@ class String(State):
         self.nodes.append(char)
 
 
+class Atom(State):
+    """
+    This state parses integer constants, boolean constants, and symbols
+    """
+
+    def __init__(self, machine):
+        State.__init__(self, machine)
+        self.initial_buf = ''
+
+    def enter(self):
+        self.buf = self.initial_buf
+
+    def exit(self):
+        self.result = _resolve_atom(self.buf)
+
+    def process(self, char):
+        """
+        State transitions:
+
+            - WHITESPACE - Idle
+            - ; - Comment
+        """
+
+        if char in WHITESPACE:
+            return Idle
+
+        if char == ";":
+            return Comment
+
+        self.buf += char
+
+
+def AtomStartingWith(initial_char):
+    def AtomFactory(machine):
+        state = Atom(machine)
+        state.initial_buf = initial_char
+        return state
+    return AtomFactory
+
+
 class Idle(State):
     """
     Idle state. This is the first (and last) thing that we should
@@ -266,7 +324,12 @@ class Idle(State):
         State transitions:
 
             - ( - Expression
-            - (default) - Error
+            - [ - List
+            - { - Dict
+            - \" - String
+            - ; - Comment
+            - # - Hash
+            - (default) - Atom
         """
 
         if char == "(":
@@ -278,6 +341,9 @@ class Idle(State):
         if char == "{":
             return Dict
 
+        if char == "\"":
+            return String
+
         if char == ";":
             return Comment
 
@@ -287,7 +353,7 @@ class Idle(State):
         if char in WHITESPACE:
             return
 
-        raise LexException("Unknown char (Idle state): `%s`" % (char))
+        return AtomStartingWith(char)
 
 
 class Comment(State):
